import datetime
import copy
import concurrent.futures
from functools import partial
from collections import OrderedDict
from hestia_earth.schema import TermJSONLD, ImpactAssessmentJSONLD, SchemaType, IndicatorJSONLD
from hestia_earth.utils.api import download_hestia

from .version import VERSION
from .representation import DataStore
from .calculation_hierarchies import Calculation_hierarchy
from .impacts.co2EqGwp100ExcludingClimateCarbonFeedbacksIpcc2013 import\
    Co2EqGwp100ExcludingClimateCarbonFeedbacksIpcc2013
from .impacts.co2EqGwp100IncludingClimateCarbonFeedbacksIpcc2013 import\
    Co2EqGwp100IncludingClimateCarbonFeedbacksIpcc2013
from .impacts.so2EqTerrestrialAcidificationIncludingFateAverageEuropeCml2001Baseline import\
    So2EqTerrestrialAcidificationIncludingFateAverageEuropeCml2001Baseline
from .impacts.so2EqTerrestrialAcidificationExcludingFateCml2001NonBaseline import\
    So2EqTerrestrialAcidificationExcludingFateCml2001NonBaseline
from .impacts.noxeqEutrophicationIncludingFateAverageEuropeCml2001NonBaseline import\
    NoxeqEutrophicationIncludingFateAverageEuropeCml2001NonBaseline
from .impacts.po4EqEutrophicationExcludingFateCml2001Baseline import Po4_EqEutrophicationExcludingFateCml2001Baseline
from .emissionsResourceUses.freshwaterWithdrawals import FreshwaterWithdrawals
from .emissionsResourceUses.identity import Identity


class Assessments_hierarchy(Calculation_hierarchy):
    """Assessments_hierarchy class. This is an abstract class from which inherit all the hierarchies defined in Hestia.
    It contains a set of common impacts and instance variables that are typically used by every hierarchy. A hierarchy
    defines the order to run the calculations of impacts in Hestia. The order is from higher tier to lower tier."""

    def __init__(self):
        """Constructor method.
            Arguments:
                None
            Returns instance."""

        super().__init__()

        self.impacts_models = [
            Co2EqGwp100ExcludingClimateCarbonFeedbacksIpcc2013(),
            Co2EqGwp100IncludingClimateCarbonFeedbacksIpcc2013(),
            Po4_EqEutrophicationExcludingFateCml2001Baseline(),
            NoxeqEutrophicationIncludingFateAverageEuropeCml2001NonBaseline(),
            So2EqTerrestrialAcidificationExcludingFateCml2001NonBaseline(),
            So2EqTerrestrialAcidificationIncludingFateAverageEuropeCml2001Baseline(),
        ]

        self.resourceUse_models = [
            Identity(),
            FreshwaterWithdrawals()
        ]

        self.calc_data = DataStore()
        self.impact_assessments = []
        self.counter = 0

    def process_hierarchy(self, data):
        """Method that runs the hierarchy through all the impacts for a given cycle.
            Arguments:
                cycle: This is a cycle in the representation required by the calculation engine.
                (Different from the Schema representation)
            Returns a file object."""

        if data.representation_cycle is not None:
            self.process_hierarchy_with_cycle(data)
        else:
            self.process_hierarchy_no_cycle(data)

    def process_hierarchy_no_cycle(self, data):

        products = [assessment['product'] for assessment in data.representation_assessment.evalues()]
        emissionsResourceUses = [assessment['emissionsResourceUse'] for assessment in
                                 data.representation_assessment.evalues()]

        for (product, emissionsResourceUse) in zip(products, emissionsResourceUses):
            prodnorm = copy.deepcopy(product)
            if 'term' in prodnorm and 'defaultProperties' in prodnorm['term']:
                prodnorm['term']['defaultProperties'] = product['term']['defaultProperties'].evalues()
            impact_assessment = self.build_impact_assessment(None, prodnorm)
            indicators = self.calculate_impact_indicators(None, prodnorm, emissionsResourceUse.evalues())
            impact_assessment['emissionsResourceUse'] = emissionsResourceUse.evalues()
            impact_assessment['impacts'] = indicators

            self.impact_assessments.append(impact_assessment)

        self.calc_data.import_assessment(self.impact_assessments)
        data.update_representation_assessments(self.calc_data)

    def process_hierarchy_with_cycle(self, data):

        cycle = data.representation_cycle
        ids_products_to_process = [assessment['product']['@id']
                                   for assessment in data.representation_assessment.evalues()]
        products_to_process = cycle['products'].evalues() if len(ids_products_to_process) == 0 else\
            [product for product in cycle['products'].evalues() if product['term']['@id'] in ids_products_to_process]

        for product in products_to_process:
            if product['economicValueShare'] != {}:
                prodnorm = copy.deepcopy(product)
                prodnorm['term']['defaultProperties'] = product['term']['defaultProperties'].evalues()
                impact_assessment = self.build_impact_assessment(cycle, prodnorm)
                emissionsResourceUses = self.calculate_emissionsResourceUses(cycle, prodnorm)
                impact_assessment['emissionsResourceUse'] = emissionsResourceUses
                indicators = self.calculate_impact_indicators(cycle, prodnorm, emissionsResourceUses)
                impact_assessment['impacts'] = indicators

                self.impact_assessments.append(impact_assessment)

        self.calc_data.import_assessment(self.impact_assessments)
        data.update_representation_assessments(self.calc_data)

    def calculate_emissionsResourceUses(self, cycle, product):

        emissionsResourceUses = {}

        for model in self.resourceUse_models:
            if not model.check(cycle, product):
                continue
            else:
                emissionsResourceUses.update(model.calculate())

        return self.build_indicators_array(emissionsResourceUses)

    def calculate_impact_indicators(self, cycle, product, emissionsResourceUses):

        results = {}
        for model in self.impacts_models:
            if not model.check(cycle, product, emissionsResourceUses):
                continue
            else:
                result = model.calculate()
                if result > 0:
                    results[model.id] = result

        return self.build_indicators_array(results)

    def build_impact_assessment(self, cycle, product):

        impact_assessment = OrderedDict(list(OrderedDict([('@context', None)]).items()) +
                                        list(ImpactAssessmentJSONLD().fields.items()))

        impact_assessment['@context'] = Calculation_hierarchy.context + "/ImpactAssessment.jsonld"
        impact_assessment['@id'] = self.counter
        impact_assessment['@type'] = SchemaType.IMPACTASSESSMENT.value
        impact_assessment['name'] = self.get_name(cycle) if cycle is not None else ''
        impact_assessment['version'] = VERSION
        impact_assessment['cycle'] = self.get_reference(cycle) if cycle is not None else {}
        impact_assessment['endDate'] = datetime.datetime.today().strftime('%Y-%m-%d')
        impact_assessment['country'] = cycle['site']['country'] if cycle is not None else ''
        impact_assessment['product'] = product['term'] if 'term' in product else product
        impact_assessment['source'] = self.get_reference(dict(cycle['defaultSource'])) if cycle is not None else {}
        impact_assessment['functionalUnitMeasure'] = "kg"
        impact_assessment['functionalUnitQuantity'] = 1
        impact_assessment['allocationMethod'] = "economic"
        impact_assessment['systemBoundary'] = True
        impact_assessment['dataPrivate'] = "false"

        impact_assessment['dataPrivate'] = False

        self.counter = self.counter + 1

        return impact_assessment

    def build_indicator(self, id, uses):
        result = uses[id]
        property = IndicatorJSONLD().fields
        term = download_hestia(id)
        term.pop('@context', None)
        property['term'] = term
        property['value'] = self.round_to_significant(result, 10)
        property['dataState'] = "recalculated"
        property['dataVersion'] = VERSION

        return property

    def build_indicators_array(self, dictionary):

        non_zero_eru = {x: y for x, y in dictionary.items() if y != 0}
        func = partial(self.build_indicator, uses=dictionary)
        with concurrent.futures.ThreadPoolExecutor() as executor:
            indicators = list(executor.map(func, list(non_zero_eru.keys())))

        return indicators

    def get_reference(self, node):

        reference = OrderedDict(list(TermJSONLD().fields.items()))
        reference['@id'] = node['@id'] if '@id' in node else None
        reference['@type'] = node['@type'] if '@type' in node else None
        reference['name'] = node['name'] if 'name' in node else None

        return reference

    def get_name(self, cycle):

        name = cycle['products'].evalues()[0]['term']['name']
        description = cycle['description'].lower() if 'description' in cycle else ""
        year = cycle['endDate'] if 'endDate' in cycle['endDate'] else ""
        source = cycle['defaultSource']['name'] if 'name' in cycle['defaultSource'] else ""

        impact_assessment_name = \
            "Impact Assessment for, '" + name + "' , under " + description + " in " + year + " from " + source

        return impact_assessment_name

    def build_emissionResourceUses(self, model):

        nodes = []
        for emissionsResourceUse in model.emissionsResourceUses:
            nodes.append(self.build_indicator(emissionsResourceUse, model.emissionsResourceUses[emissionsResourceUse]))

        return nodes
