import math
from collections import OrderedDict
from hestia_earth.calculation.abstract_hierarchy import Hierarchy
from hestia_earth.schema import TermJSONLD, EmissionJSONLD, EmissionDataState, SchemaType
from .version import VERSION


class Calculation_hierarchy(Hierarchy):
    """Model_Hierarchy class. This is an abstract class from which inherit all the hierarchies defined in Hestia. It
    contains a set of common methods and instance variables that are typically used by every hierarchy. A hierarchy
    defines the order to run the calculations in the models for a particular emission defined in Hestia. The order
    is from higher tier to lower tier."""

    def __init__(self):
        """Constructor method.
            Arguments:
                None
            Returns instance."""

        super().__init__()

        self.models = []

    def process_hierarchy(self, data):
        """Method that runs the hierarchy through all the models for a given cycle.
            Arguments:
                cycle: This is a cycle in the representation required by the calculation engine.
                (Different from the Schema representation)
            Returns a file object."""

        cycle = data.representation_cycle

        for model in self.models:
            if not self.check_model(model, cycle):
                continue
            else:
                self.chosen_model = model
                self.calculate_model(model, cycle)
                self.update_node(data)
                break

    def check_model(self, model, cycle):
        """Method that checks if the model has enough data to do the calculation.
            Arguments:
                cycle: This is a cycle in the representation required by the calculation engine.
                (Different from the Schema representation)
            Returns boolean."""

        pass

    def calculate_model(self, model, cycle):
        """Method that calculates the data if the model has enough data to do the calculation.
            Arguments:
                cycle: This is a cycle in the representation required by the calculation engine.
                (Different from the Schema representation)
            Returns boolean."""

        pass

    def update_node(self, data):
        """Method that updates cycle representation.
            Arguments:
                data: It is a dataStore object with the representation
            Returns boolean."""

        pass

    def round_to_significant(self, number, sig_number):
        """Method to round to a given number of significant numbers.
            Arguments:
                number: The number to round.
                sig_number: The number of significant numbers.
            Returns The rounded number."""
        return round(number, sig_number - int(math.floor(math.log10(abs(number)))) - 1) if number > 0 else 0


class Emissions_hierarchy(Calculation_hierarchy):
    """Model_Hierarchy class. This is an abstract class from which inherit all the hierarchies defined in Hestia. It
    contains a set of common methods and instance variables that are typically used by every hierarchy. A hierarchy
    defines the order to run the calculations in the models for a particular emission defined in Hestia. The order
    is from higher tier to lower tier."""

    def __init__(self):
        """Constructor method.
            Arguments:
                None
            Returns instance."""

        super().__init__()
        self.emission = None

    def build_emission(self):

        self.emission = EmissionJSONLD()
        self.emission.fields['@type'] = SchemaType.EMISSION.value
        self.emission.fields['term'] = dict(self.get_reference())
        self.emission.fields['description'] = self.term['description'] if self.term.get('description') else None
        self.emission.fields['value'] = [self.round_to_significant(self.best_result, 10)]
        self.emission.fields['dataState'] = EmissionDataState.RECALCULATED.value
        if self.best_tier != "background":
            self.emission.fields['methodTier'] = f"tier {self.best_tier}"
        else:
            self.emission.fields['methodTier'] = self.best_tier
            self.emission.fields['inputs'] = self.chosen_model.assessment_inputs

        self.emission.fields['dataVersion'] = VERSION

        return dict(self.emission.to_dict())

    def get_reference(self):

        reference = TermJSONLD()

        reference.fields = OrderedDict(list(reference.fields.items()))
        reference.fields["@id"] = self.term["@id"]
        reference.fields["@type"] = self.term["@type"]
        reference.fields["name"] = self.term["name"]
        reference.fields["units"] = self.term["units"]

        del(reference.fields['termType'])

        return reference.to_dict()

    def update_node(self, data):
        """Method that updates cycle representation emission.
            Arguments:
                data: It is a dataStore object with the representation with the emission to be updated
            Returns boolean."""

        data.update_representation_emission(self.build_emission())


class Impacts_hierarchy(Calculation_hierarchy):
    """Model_Hierarchy class. This is an abstract class from which inherit all the hierarchies defined in Hestia. It
    contains a set of common methods and instance variables that are typically used by every hierarchy. A hierarchy
    defines the order to run the calculations in the models for a particular emission defined in Hestia. The order
    is from higher tier to lower tier."""

    def __init__(self):
        """Constructor method.
            Arguments:
                None
            Returns instance."""

        super().__init__()

    def update_representation(self, model, cycle):
        """Method that checks if the model has enough data to do the calculation.
            Arguments:
                cycle: This is a cycle in the representation required by the calculation engine.
                (Different from the Schema representation)
            Returns boolean."""

        cycle['impacts'][self.term['@id']] = model.emision
