from hestia_earth.calculation.abstract_model import Model
from hestia_earth.utils.lookup import load_lookup, get_table_value


class BiodiversityImpactLandUseChaudhary2015(Model):
    def __init__(self):

        # Instantiate variables
        self.id = "speciesDestinedToExtinctionPerYearBiodiversityLossPermanentEffectsLc-Impact"
        self.transBioImpact = None
        self.occBioImpact = None
        self.LandUseChange = None
        self.lut_1 = None
        self.loc_1 = None

        # look up tables
        self.look_up_global_occCF = load_lookup(
            "hestia_earth/calculation/data/lookups/biodiversity_cf_chaudhary_2015_global_occupational.csv")
        self.look_up_global_transCF = load_lookup(
            "hestia_earth/calculation/data/lookups/biodiversity_cf_chaudhary_2015_global_transformation.csv")
        self.look_up_biodiversity_country_LUC = load_lookup(
            "hestia_earth/calculation/data/lookups/Biodiversity_LandUseChange_PercentPerYear_BlonkConsultants2014.csv")

    def calculate(self, cycle, product, emissionsResourceUse):
        # TODO: How to deal with products with multiple land use types for a single product, for example, beef.
        # Calculate biodiversity impacts (Chaudhary 2015)

        # 1) Estimate land occupation biodiversity impact - Potential global species loss due to land use occupation

        # Identify the broad land use category (cropGrouping) based on crop type
        crop_properties = product.get('defaultProperties', [])
        crop_groups = [cg.get('value', []) for cg in crop_properties if cg.get('term', {}).get('@id', '')
                       == 'cropGroupings']

        crop_groups_list = crop_groups[0].split("; ")

        self.LU_type = list(set(crop_groups_list).intersection(['Annual crops', 'Permanent crops', 'Pasture']))
        # TODO: Add a catch to break if greater than 1 land use type here
        self.LU_type[0] = self.LU_type[0].replace(" ", "_")  # Replace blank with underscore to match lookup data
        self.LU_type[0] = self.LU_type[0].lower()  # refactor this

        # Extract the full country name based on the three character gadm country name from the cycle
        # use ecoregion if available, else use country
        if cycle['site']['ecoregion'] != {}:
            self.loc_1 = cycle['site']['ecoregion']['@id'].replace('ecoregion-', '')
        else:
            self.loc_1 = cycle['site']['country']['name']

        # TODO: ensure names are consistent between region glossary and the biodiversity CF lookups
        # self.loc_1 = self.look_up_country_match.loc[self.look_up_country_match['gadmCountry'] ==
        #                                            cycle["site"]["country"], 'fullCountry']

        # Pull the occCF for the country of interest for the land use in question
        # TODO: Update to cover specific taxonomic groups (for exmaple, mammals, birds, plants...)
        self.LU_column_name = self.LU_type[0] + '_taxa_aggregated_median'
        self.temp_occCF = float(get_table_value(self.look_up_global_occCF, 'region_code', self.loc_1,
                                                self.LU_column_name))

        # Extract the landOccupation value
        emissionsResourceUses = cycle.get('emissionsResourceUse', [])
        self.land_occupation_dict = [eru for eru in emissionsResourceUses if eru.get('term', {}).get('@id', '')
                                     == 'landOccupation']

        self.land_occupation = self.land_occupation_dict[0]['value']

        # occBioImpact = occCF * landOccupation (m2*year of land needed to produce 1kg of the product)
        self.occBioImpact = self.land_occupation * self.temp_occCF

        # 2) Estimate land transformation biodiversity impact - Pot. glob. species loss due to land use transformation
        # Pull the broad crop data from the glossary #
        self.lut_1 = self.LU_type[0].replace("_", "")

        # self.lut_1 = self.lut_1.replace(" ", ".") # reformat to match the column header in the LUC lookup
        # TODO: Allow users to specify land use change for a given production cycle rather than taking a country
        #  average (for the crop in question) as is done here

        # Estimate the amount of land transformed for the given product. CHECK Divide by 100 to convert to proportion
        self.LandUseChange_percent = float(get_table_value(self.look_up_biodiversity_country_LUC, 'country',
                                                           cycle['site']['country']['name'].lower(), self.lut_1))
        self.LandUseChange = self.LandUseChange_percent/100

        # TODO: Pasture land use change calculated as sum of temp. and perm. pasture land use change percent

        # TODO: Update country and crop names within this lookup table to match the GoT

        # TODO: Discuss with JP shoudl we use the actual crop type here? for example, Cassava is convtered to the
        #  default "annual crop %" rather than "Cassava %" - Data limitation issue?

        # Pull the transCF for the region and land use in question
        self.temp_transCF = float(get_table_value(self.look_up_global_transCF, 'region_code', self.loc_1,
                                                  self.LU_column_name))

        # self.transCF_region = (landOccupation * proportion of land converted) * temp_transCF
        # Estimate transformation impact.
        self.transBioImpact = self.land_occupation * self.LandUseChange * self.temp_transCF

        # 3) Estimate potential global species loss due to land use impacts.
        biodiversityImpact = self.occBioImpact + self.transBioImpact  # Sum the two impacts to calculaet the overall
        # biodiversity impact

        # return [self.biodiversity_M1, self.occCF_region, self.transCF_region]
        # create emissions resource use node
        # impacts = {}

        # impacts[
        #    'speciesDestinedToExtinctionPerYearBiodiversityLossPermanentEffectsLc-Impact'] = biodiversityImpact
        # emissionsResourceUses[
        #    'speciesDestinedToExtinctionPerYearBiodiversityLossLandOccupationLc-Impact'] = occBioImpact
        # emissionsResourceUses[
        #    'speciesDestinedToExtinctionPerYearBiodiversityLossLandTransformationLc-Impact'] = transBioImpact
        return biodiversityImpact  # return the biodiversity impact value

    def check(self, cycle, product, emissionsResourceUse):
        # extract landOccupation - if it exists
        emissionsResourceUses = cycle.get('emissionsResourceUse', [])
        land_occupation_included = [eru for eru in emissionsResourceUses if eru.get('term', {}).get('@id', '')
                                    == 'landOccupation']

        # extract broad cropGrouping - if it exists
        crop_properties = product.get('defaultProperties', [])
        crop_groups = [cg.get('value', []) for cg in crop_properties if cg.get('term', {}).get('@id', '')
                       == 'cropGroupings']
        crop_groups_list = crop_groups[0].split("; ")
        LandUseIncluded = list(set(crop_groups_list).intersection(['Annual crops', 'Permanent crops', 'Pasture']))

        # if we have landOccupation data and the product has an EVS > 0 then pass the check
        if land_occupation_included != [] and product.get('economicValueShare', 0) > 0 and LandUseIncluded != []:
            return True
        else:
            return False
