from hestia_earth.calculation.abstract_model import Model
from hestia_earth.calculation.data.constants.nh3 import NH3_FERTILIZER_FACTOR
from hestia_earth.calculation.utils import most_relevant_measurement


class NH3ToAirInorganicFertilizerEuropeanEnvironmentAgency2016(Model):
    def __init__(self):

        # Define model tier
        self.tier = 1

        # Define model requirements
        self.soilPh = None
        self.temperature = None
        self.ureaAsN = None
        self.ureaAmmoniumSulphateAsN = None
        self.ammoniumBicarbonateAsN = None
        self.ammoniumSulphateAsN = None
        self.ureaAmmoniumNitrateAsN = None
        self.ammoniumNitrateAsN = None
        self.ammoniumChlorideAsN = None
        self.potassiumNitrateAsN = None
        self.potassiumNitrateAsK2O = None
        self.calciumAmmoniumNitrateAsN = None
        self.anhydrousAmmoniumAsN = None
        self.diammoniumPhosphateAsN = None
        self.monoammoniumPhosphateAsP2O5 = None

        # Instantiate variables
        self.nh3ToAirInorganicFertilizer = None

        # Define model coeffients
        self.emf_lookup = NH3_FERTILIZER_FACTOR

    def calculate_nh3ToAirInorganicFertilizer(self):
        # Calculate total nh3 emissions

        emfs = self.get_emissions_mineral_fertilizer(self.soilPh, self.temperature)

        ureaAsN = self.ureaAsN if self.ureaAsN != {} else 0
        ureaAmmoniumSulphateAsN = self.ureaAmmoniumSulphateAsN if self.ureaAmmoniumSulphateAsN != {} else [0]
        ammoniumBicarbonateAsN = self.ammoniumBicarbonateAsN if self.ammoniumBicarbonateAsN != {} else [0]
        ammoniumSulphateAsN = self.ammoniumSulphateAsN if self.ammoniumSulphateAsN != {} else [0]
        ureaAmmoniumNitrateAsN = self.ureaAmmoniumNitrateAsN if self.ureaAmmoniumNitrateAsN != {} else [0]
        ammoniumNitrateAsN = self.ammoniumNitrateAsN if self.ammoniumNitrateAsN != {} else [0]
        ammoniumChlorideAsN = self.ammoniumNitrateAsN if self.ammoniumNitrateAsN != {} else [0]
        potassiumNitrateAsN = self.potassiumNitrateAsN if self.potassiumNitrateAsN != {} else [0]
        potassiumNitrateAsK2O = self.potassiumNitrateAsK2O if self.potassiumNitrateAsK2O != {} else [0]
        calciumAmmoniumNitrateAsN = self.calciumAmmoniumNitrateAsN if self.calciumAmmoniumNitrateAsN != {} else [0]
        anhydrousAmmoniumAsN = self.anhydrousAmmoniumAsN if self.anhydrousAmmoniumAsN != {} else [0]
        diammoniumPhosphateAsN = self.diammoniumPhosphateAsN if self.diammoniumPhosphateAsN != {} else [0]
        monoammoniumPhosphateAsP2O5 = self.monoammoniumPhosphateAsP2O5 if self.monoammoniumPhosphateAsP2O5 != {} \
            else [0]

        Urea_UAS_AmmBicarb = ureaAsN + ureaAmmoniumSulphateAsN + ammoniumBicarbonateAsN

        AN_ACl_NP_NPK = ammoniumNitrateAsN + ammoniumChlorideAsN + potassiumNitrateAsN + potassiumNitrateAsK2O

        AP = diammoniumPhosphateAsN + monoammoniumPhosphateAsP2O5

        n_inputs = [sum(Urea_UAS_AmmBicarb), sum(ammoniumSulphateAsN),
                    sum(ureaAmmoniumNitrateAsN), sum(AN_ACl_NP_NPK),
                    sum(calciumAmmoniumNitrateAsN),
                    sum(anhydrousAmmoniumAsN), sum(AP)]

        self.nh3ToAirInorganicFertilizer = sum([x*y for x, y in zip(n_inputs, emfs)])

        return self.nh3ToAirInorganicFertilizer

    def get_emissions_mineral_fertilizer(self, ph, temperature):

        conditions = [ph <= 7 and temperature <= 14,
                      ph <= 7 and temperature > 14 and temperature < 26,
                      ph <= 7 and temperature >= 26,
                      ph > 7 and temperature <= 14,
                      ph > 7 and temperature > 14 and temperature < 26,
                      ph > 7 and temperature >= 26]

        factors = [self.emf_lookup['soil ph < 7']['cool T < 14°C'].values(),
                   self.emf_lookup['soil ph < 7']['room temperature  15°C < T < 25°C'].values(),
                   self.emf_lookup['soil ph < 7']['warm T >= 26°C'].values(),
                   self.emf_lookup['soil ph >= 7']['cool T < 14°C'].values(),
                   self.emf_lookup['soil ph >= 7']['room temperature  15°C < T < 25°C'].values(),
                   self.emf_lookup['soil ph >= 7']['warm T >= 26°C'].values()]

        return factors[conditions.index(True)]

    def check_nh3ToAirInorganicFertilizer(self, cycle):
        # Check that we have all the inputs

        self.temperature =\
            most_relevant_measurement(cycle['site']['measurements']['annualMeanTemperatureLong-TermMean'],
                                      cycle['endDate'])
        self.soilPh = most_relevant_measurement(cycle['site']['measurements']['soilPh'], cycle['endDate'])
        self.ureaAsN = cycle['inputs']['ureaAsN']['value']
        self.ureaAmmoniumSulphateAsN = cycle['inputs']['ureaAmmoniumSulphateAsN']['value']
        self.ammoniumBicarbonateAsN = cycle['inputs']['ammoniumBicarbonateAsN']['value']
        self.ammoniumSulphateAsN = cycle['inputs']['ammoniumSulphateAsN']['value']
        self.ureaAmmoniumNitrateAsN = cycle['inputs']['ureaAmmoniumNitrateAsN']['value']
        self.ammoniumNitrateAsN = cycle['inputs']['ammoniumNitrateAsN']['value']
        self.ammoniumChlorideAsN = cycle['inputs']['ammoniumChlorideAsN']['value']
        self.potassiumNitrateAsN = cycle['inputs']['potassiumNitrateAsN']['value']
        self.potassiumNitrateAsK2O = cycle['inputs']['potassiumNitrateAsK2O']['value']
        self.calciumAmmoniumNitrateAsN = cycle['inputs']['calciumAmmoniumNitrateAsN']['value']
        self.anhydrousAmmoniumAsN = cycle['inputs']['anhydrousAmmoniumAsN']['value']
        self.diammoniumPhosphateAsN = cycle['inputs']['diammoniumPhosphateAsN']['value']
        self.monoammoniumPhosphateAsP2O5 = cycle['inputs']['monoammoniumPhosphateAsP2O5']['value']

        checkin = [self.ureaAsN, self.ureaAmmoniumSulphateAsN, self.ammoniumSulphateAsN, self.ureaAmmoniumNitrateAsN,
                   self.ammoniumNitrateAsN, self.ammoniumChlorideAsN, self.potassiumNitrateAsN,
                   self.potassiumNitrateAsK2O, self.calciumAmmoniumNitrateAsN, self.anhydrousAmmoniumAsN,
                   self.diammoniumPhosphateAsN, self.monoammoniumPhosphateAsP2O5]

        return any(v != {} for v in checkin) and self.temperature != {} and self.soilPh != {}
