from hestia_earth.calculation.abstract_model import Model
from hestia_earth.calculation.utils import most_relevant_measurement, summation
from hestia_earth.calculation.data.constants.no3 import NO3_LEECHING_FACTORS
from hestia_earth.calculation.data.constants.generic import ATOMIC_WEIGHT_CONVERSIONS


class NO3ToGroundwaterAllInputsPooreNemecek2018(Model):
    def __init__(self):

        # Define model tier
        self.tier = 2

        # Define model requirements
        self.rootingDepth = None
        self.clay = None
        self.sand = None
        self.precipitation = None
        self.inorgN_total = None
        self.orgN_total = None
        self.excretaN_total = None
        self.res_nitrogen = None
        self.N_total = None

        # Instantiate variables
        self.no3ToGroundwaterAllOrigins = None
        self.no3ToGroundwaterOrganicFertilizer = None
        self.no3ToGroundwaterInorganicFertilizer = None
        self.no3ToGroundwaterExcreta = None
        self.no3ToGroundwaterCropResidueDecomposition = None

        # Define model coefficients
        self.leaching = NO3_LEECHING_FACTORS
        self.conv_no3n_no3 = ATOMIC_WEIGHT_CONVERSIONS['Conv_Mol_NO3N_NO3']

    def calculate_no3ToGroundwaterInorganicFertilizer(self):

        self.no3ToGroundwaterInorganicFertilizer = \
            self.calculate_no3ToGroundwaterAllOrigins() * self.inorgN_total / self.N_total

        return self.no3ToGroundwaterInorganicFertilizer

    def calculate_no3ToGroundwaterOrganicFertilizer(self):

        self.no3ToGroundwaterOrganicFertilizer = \
            self.calculate_no3ToGroundwaterAllOrigins() * self.orgN_total / self.N_total

        return self.no3ToGroundwaterOrganicFertilizer

    def calculate_no3ToGroundwaterExcreta(self):

        self.no3ToGroundwaterExcreta = self.calculate_no3ToGroundwaterAllOrigins() * self.excretaN_total / self.N_total

        return self.no3ToGroundwaterExcreta

    def calculate_no3ToGroundwaterCropResidueDecomposition(self):

        self.no3ToGroundwaterCropResidueDecomposition = \
            self.calculate_no3ToGroundwaterAllOrigins() * self.res_nitrogen / self.N_total

        return self.no3ToGroundwaterCropResidueDecomposition

    def calculate_no3ToGroundwaterAllOrigins(self):

        self.no3ToGroundwaterAllOrigins = self.N_total * self.get_leaching_factor() * self.conv_no3n_no3

        return self.no3ToGroundwaterAllOrigins

    def get_res_n(self, cycle):

        AG_Res = cycle['products']['aboveGroundCropResidueDryMatter']['value']
        AG_Res_ncontent = cycle['products']['aboveGroundCropResidueDryMatter']['properties']['nitrogenContent']['value']
        AG_Res_N = sum(AG_Res) * float(AG_Res_ncontent)/100 if AG_Res != {} and AG_Res_ncontent != {} else {}

        BG_Res = cycle['products']['belowGroundCropResidueDryMatter']['value']
        BG_Res_ncontent = cycle['products']['belowGroundCropResidueDryMatter']['properties']['nitrogenContent']['value']
        BG_Res_N = sum(BG_Res) * float(BG_Res_ncontent)/100 if BG_Res != {} and BG_Res_ncontent != {} else {}

        N_store = [AG_Res_N, BG_Res_N]

        nitrogen_store = [element for element in N_store if element != {}]

        return nitrogen_store

    def get_total_n(self, cycle):

        inputs = cycle['inputs'].evalues()

        N_total = [sum(input['value']) if 'units' in input['term'] and input['term']['units'] == 'kg N'
                   else None for input in inputs]

        N_total = [element for element in N_total if element is not None]

        return N_total

    def get_leaching_factor(self):
        if (self.rootingDepth > 1.3 or self.clay > 0.50 or self.precipitation < 500) and\
                (self.rootingDepth > 0.4 or self.sand < 0.85 or self.precipitation < 1300):
            return self.leaching['low']
        elif (self.rootingDepth < 0.4 or self.sand > 0.85 or self.precipitation > 1300) and\
                (self.rootingDepth > 1.3 or self.clay < 0.5 or self.precipitation > 500):
            return self.leaching['high']
        else:
            return self.leaching['other']

    def check_no3ToGroundwaterInorganicFertilizer(self, cycle):

        inputs = cycle['inputs'].evalues()

        inorgN_filter = [sum(input['value']) if 'units' in input['term'] and input['term']['units'] == 'kg N' and
                         input['term']['termType'] == 'inorganicFertilizer'
                         else None for input in inputs]

        self.inorgN_total = summation([element for element in inorgN_filter if element is not None])

        return self.inorgN_total != {} and self.check_no3ToGroundwaterAllOrigins(cycle) and self.N_total != 0

    def check_no3ToGroundwaterOrganicFertilizer(self, cycle):

        inputs = cycle['inputs'].evalues()

        orgN_filter = [sum(input['value']) if 'units' in input['term'] and
                                              input['term']['units'] == 'kg N' and
                                              input['term']['termType'] == 'organicFertilizer' else None
                       for input in inputs]

        orgN_filter = [element for element in orgN_filter if element is not None]

        self.orgN_total = sum(orgN_filter)

        return len(orgN_filter) > 0 and self.check_no3ToGroundwaterAllOrigins(cycle) and self.N_total != 0

    def check_no3ToGroundwaterExcreta(self, cycle):

        inputs = cycle['inputs'].evalues()

        excretaN_filter = [sum(input['value']) if 'units' in input['term'] and input['term']['units'] == 'kg N'
                                                  and input['term']['termType'] == 'animalProduct' else None
                           for input in inputs]

        self.excretaN_total = summation([element for element in excretaN_filter if element is not None])

        return self.excretaN_total != {} and self.check_no3ToGroundwaterAllOrigins(cycle) and self.N_total != 0

    def check_no3ToGroundwaterCropResidueDecomposition(self, cycle):

        self.res_nitrogen = summation(self.get_res_n(cycle))

        return self.res_nitrogen != {} and self.check_no3ToGroundwaterAllOrigins(cycle) and self.N_total != 0

    def check_no3ToGroundwaterAllOrigins(self, cycle):

        rootingDepth_vector = [float(product['properties']['rootingDepth']['value']) for product in
                               cycle['products'].evalues() if 'properties' in product
                               and 'rootingDepth' in product['properties']]
        self.rootingDepth = {} if rootingDepth_vector == [] else sum(rootingDepth_vector)

        clay = most_relevant_measurement(cycle['site']['measurements']['clayContent'], cycle['endDate'])
        sand = most_relevant_measurement(cycle['site']['measurements']['sandContent'], cycle['endDate'])

        self.clay = clay/100 if clay != {} else {}
        self.sand = sand/100 if sand != {} else {}

        cycle_rainfall = most_relevant_measurement(cycle['site']['measurements']['annualRainfallYearOfCycle'],
                                                   cycle['endDate'])
        annualrainfall = most_relevant_measurement(cycle['site']['measurements']['annualRainfallLong-TermMean'],
                                                   cycle['endDate'])
        self.precipitation = cycle_rainfall if cycle_rainfall != {} else annualrainfall if annualrainfall != {} else {}

        self.N_total = summation(self.get_total_n(cycle) + self.get_res_n(cycle))

        return self.N_total != {} and self.rootingDepth != {} and self.clay != {} and self.sand != {} and\
            self.precipitation != {}
