from hestia_earth.calculation.abstract_model import Model
from hestia_earth.calculation.utils import most_relevant_measurement, summation
from hestia_earth.calculation.data.constants.nox import NOX_FACTORS_BY_CLIMATE_ZONE, DRY_MATTER_FACTOR_TO_NOX
from hestia_earth.calculation.data.constants.generic import ATOMIC_WEIGHT_CONVERSIONS
import numpy as np


class NOXToAirFertilizerAndExcretaStehfestBouwman2006(Model):
    def __init__(self):

        # Define model tier
        self.tier = 2

        # Define model requirements
        self.nitrogenContent = None
        self.eco_ClimateZone = None
        self.crop_name = None
        self.inorgN_total = None
        self.orgN_total = None
        self.excretaN_total = None
        self.res_nitrogen = None
        self.N_total = None

        # Instantiate variables
        self.noxToAirInorganicFertilizer = None
        self.noxToAirOrganicFertilizer = None
        self.noxToAirExcreta = None
        self.noxToAirCropResidueDecomposition = None
        self.noxToAirCropResidueBurning = None
        self.noxToAirAllOrigins = None

        # Define model coefficients
        self.crop_nox_n = NOX_FACTORS_BY_CLIMATE_ZONE
        self.conv_non_no = ATOMIC_WEIGHT_CONVERSIONS['Conv_Mol_NON_NO']
        self.drymatter_to_nox = DRY_MATTER_FACTOR_TO_NOX

    def calculate_noxToAirInorganicFertilizer(self):

        self.noxToAirInorganicFertilizer = \
            self.calculate_noxToAirAllOrigins() * self.inorgN_total / self.N_total

        return self.noxToAirInorganicFertilizer

    def calculate_noxToAirOrganicFertilizer(self):

        self.noxToAirOrganicFertilizer = self.calculate_noxToAirAllOrigins() * self.orgN_total / self.N_total

        return self.noxToAirOrganicFertilizer

    def calculate_noxToAirExcreta(self):

        self.noxToAirExcreta = self.calculate_noxToAirAllOrigins() * self.excretaN_total / self.N_total

        return self.noxToAirExcreta

    def calculate_noxToAirCropResidueDecomposition(self):

        self.noxToAirCropResidueDecomposition = \
            self.calculate_noxToAirAllOrigins() * self.res_nitrogen / self.N_total

        return self.noxToAirCropResidueDecomposition

    def calculate_noxToAirCropResidueBurning(self):

        self.noxToAirCropResidueBurning = self.burnt * self.drymatter_to_nox

        return self.noxToAirCropResidueBurning

    def calculate_noxToAirAllOrigins(self):

        self.noxToAirAllOrigins = min(
            0.025 * self.N_total,
            np.exp(-0.451 + 0.0061 * self.N_total + (0 if self.nitrogenContent / 1000000 < 0.0005 else
                                                     -1.0211 if self.nitrogenContent / 1000000 <= 0.002
                                                     else 0.7892) +
                   self.crop_nox_n[str(self.eco_ClimateZone)]) -
            np.exp(-0.451 + (0 if self.nitrogenContent / 1000000 < 0.0005 else
                             -1.0211 if self.nitrogenContent / 1000000 <= 0.002 else 0.7892) +
                   self.crop_nox_n[str(self.eco_ClimateZone)])) * self.conv_non_no

        return self.noxToAirAllOrigins

    def get_res_n(self, cycle):

        AG_Res = cycle['products']['aboveGroundCropResidueDryMatter']['value']
        AG_Res_ncontent = cycle['products']['aboveGroundCropResidueDryMatter']['properties']['nitrogenContent']['value']
        AG_Res_N = sum(AG_Res) * float(AG_Res_ncontent)/100 if AG_Res != {} and AG_Res_ncontent != {} else {}

        BG_Res = cycle['products']['belowGroundCropResidueDryMatter']['value']
        BG_Res_ncontent = cycle['products']['belowGroundCropResidueDryMatter']['properties']['nitrogenContent']['value']
        BG_Res_N = sum(BG_Res) * float(BG_Res_ncontent)/100 if BG_Res != {} and BG_Res_ncontent != {} else {}

        N_store = [AG_Res_N, BG_Res_N]

        nitrogen_store = [element for element in N_store if element != {}]

        return nitrogen_store

    def get_total_n(self, cycle):

        inputs = cycle["inputs"].evalues()

        N_total = [sum(input["value"]) if "units" in input["term"] and input["term"]["units"] == "kg N"
                   else None for input in inputs]

        N_total = [element for element in N_total if element is not None]

        return N_total

    def check_noxToAirInorganicFertilizer(self, cycle):

        inputs = cycle['inputs'].evalues()

        inorgN_filter = [sum(input['value']) if 'units' in input['term'] and input['term']['units'] == 'kg N' and
                         input['term']['termType'] == 'inorganicFertilizer'
                         else None for input in inputs]

        self.inorgN_total = summation([element for element in inorgN_filter if element is not None])

        return self.inorgN_total != {} and self.check_noxToAirAllOrigins(cycle) and self.N_total != 0

    def check_noxToAirOrganicFertilizer(self, cycle):

        inputs = cycle['inputs'].evalues()

        orgN_filter = [sum(input['value']) if 'units' in input['term'] and
                                              input['term']['units'] == 'kg N' and
                                              input['term']['termType'] == 'organicFertilizer' else None
                       for input in inputs]

        self.orgN_total = summation([element for element in orgN_filter if element is not None])

        return self.orgN_total != {} and self.check_noxToAirAllOrigins(cycle) and self.N_total != 0

    def check_noxToAirCropResidueDecomposition(self, cycle):

        self.res_nitrogen = summation(self.get_res_n(cycle))

        return self.res_nitrogen != {} and self.check_noxToAirAllOrigins(cycle) and self.N_total != 0

    def check_noxToAirExcreta(self, cycle):

        inputs = cycle["inputs"].evalues()

        excretaN_filter = [sum(input["value"]) if "units" in input["term"] and input["term"]["units"] == "kg N"
                                                  and input["term"]["termType"] == "animalProduct" else None
                           for input in inputs]

        self.excretaN_total = summation([element for element in excretaN_filter if element is not None])

        return self.excretaN_total != {} and self.check_noxToAirAllOrigins(cycle) and self.N_total != 0

    def check_noxToAirCropResidueBurning(self, cycle):

        filter_list = ['aboveGroundCropResidueDryMatter', 'belowGroundCropResidueDryMatter']

        products = cycle["products"].evalues()
        self.burnt = summation([sum(product['value']) for product in products
                                if product['term']['@id'] in filter_list and 'destination' in product and
                                'burning' == product['destination']['@id']])

        return self.burnt != {} and self.check_noxToAirAllOrigins(cycle) and self.N_total != 0

    def check_noxToAirAllOrigins(self, cycle):

        self.eco_ClimateZone = most_relevant_measurement(cycle['site']['measurements']['eco-ClimateZone'],
                                                         cycle['endDate'])
        self.nitrogenContent = most_relevant_measurement(cycle['site']['measurements']['soilTotalNitrogenContent'],
                                                         cycle['endDate'])
        first_product = list(cycle['products'].evalues())
        self.crop_name = first_product[0]['term']['name']
        self.N_total = summation(self.get_total_n(cycle) + self.get_res_n(cycle))

        return self.N_total != {} and self.nitrogenContent != {} and self.eco_ClimateZone != {} and self.crop_name != {}
