import numpy as np
from hestia_earth.calculation.abstract_model import Model
from hestia_earth.calculation.data.constants.p import C2_FACTORS_TILLAGE, SLOPE_RANGE
from hestia_earth.calculation.utils import most_relevant_measurement
from hestia_earth.utils.lookup import download_lookup


class PErosionSalcaSchererPfister2015(Model):
    def __init__(self):

        # Define model tier
        self.tier = 1

        # Define model requirements
        self.nLAEnvironment = None
        self.soilPhosphorusContent = None
        self.irrigation = None
        self.precipitation = None
        self.wtype = None
        self.product = None
        self.erodibility = None
        self.slope = None
        self.slopeLength = None
        self.country = None
        self.ppractice = None
        self.management = None
        self.pcorr = None

        # Instantiate variables
        self.A = None
        self.R = None
        self.pErosionAllOrigins = None

        # Define model coefficients
        self.c1 = {elem[0]: elem[6] for elem in download_lookup('crop.csv') if not np.isnan(elem[6])}
        self.c2 = {elem[0]: np.float64(elem[3]) for elem in download_lookup('region.csv')
                   if elem[3] != '-' and elem[3] != ""}
        self.tillage = C2_FACTORS_TILLAGE
        self.srange = SLOPE_RANGE

    def get_pcorr(self, slope):

        pcorr = None
        for element in self.srange:
            if slope >= element[0] and slope < element[1]:
                pcorr = element[2]
                break
            else:
                pcorr = None

        return pcorr

    def get_management(self, cycle):

        filter_management = ['fullTillage', 'reducedTillage', 'noTillage']

        management = None
        for practice in cycle['practices'].evalues():
            if practice['term']['@id'] in filter_management:
                management = practice['term']['@id']

        return management

    def calculate_R(self):

        if (self.irrigation + self.precipitation) > 850:
            R = 587.8 - 1.219 * (self.irrigation + self.precipitation) + \
                     0.004105 * (self.irrigation + self.precipitation) ** 2
        else:
            R = 0.0483 * (self.irrigation + self.precipitation) ** 1.61

        return R

    def calculate_A(self):

        A = self.R * self.erodibility * self.slopeLength * self.ppractice * self.pcorr * self.c1[self.product] * \
            (float(self.c2[self.country]) if self.product == 'Pasture' or not self.management
             else self.tillage[self.management])

        return A

    def calculate_pErosionAllOrigins(self):

        self.R = self.calculate_R()
        self.A = self.calculate_A()
        self.pErosionAllOrigins = self.A * self.nLAEnvironment/100 * 2 * self.soilPhosphorusContent/1000

        return self.pErosionAllOrigins

    def check_pErosionAllOrigins(self, cycle):

        self.nLAEnvironment =\
            most_relevant_measurement(cycle['site']['measurements']['nutrientLossToAquaticEnvironment'],
                                      cycle['endDate'])
        self.soilPhosphorusContent =\
            most_relevant_measurement(cycle['site']['measurements']['soilPhosphorusContent'], cycle['endDate'])

        self.irrigation = sum(cycle['inputs']['irrigationPumpedGroundWater']['value']) \
            if cycle['inputs']['irrigationPumpedGroundWater']['value'] != {} else {}
        self.precipitation =\
            most_relevant_measurement(cycle['site']['measurements']['annualRainfallLong-TermMean'], cycle['endDate'])
        self.wtype = most_relevant_measurement(cycle['site']['measurements']['heavyWinterPrecipitation'],
                                               cycle['endDate'])
        self.product = cycle['products'].evalues()[0]['term']['@id'] if cycle['products'] != {} else {}
        self.erodibility = most_relevant_measurement(cycle['site']['measurements']['erodibility'], cycle['endDate'])
        self.slope = most_relevant_measurement(cycle['site']['measurements']['slope'], cycle['endDate'])
        self.slopeLength = most_relevant_measurement(cycle['site']['measurements']['slopeLength'], cycle['endDate'])
        self.country = cycle['site']['country']['@id']
        self.ppractice = cycle['practices']['practiceFactor']['value']

        self.management = self.get_management(cycle)
        self.pcorr = self.get_pcorr(self.slope) if self.slope != {} else {}

        return self.nLAEnvironment != {} and self.soilPhosphorusContent != {} and self.irrigation != {} and\
            self.precipitation != {} and self.wtype != {} and self.product != {} and self.product in self.c1 and\
            self.erodibility != {} and self.slope != {} and self.slopeLength != {} and self.country != {} and\
            self.country in self.c2 and self.ppractice != {} and self.management != {} and self.pcorr != {}
