import unittest
import json
from tests.utils import fixtures_path
from hestia_earth.calculation.abstract_engine import Engine
from hestia_earth.calculation.models.noxToAirFertilizerAndExcretaStehfestBouwman2006\
    import NOXToAirFertilizerAndExcretaStehfestBouwman2006


class Test_noxToAirFertilizerAndExcretaStehfestBouwman2006(unittest.TestCase):
    def __init__(self, *args, **kwargs):

        super(Test_noxToAirFertilizerAndExcretaStehfestBouwman2006, self).__init__(*args, **kwargs)

        with open(f"{fixtures_path}/WI5hWtsqlu/Cycle/WI5hWtsqlu.jsonld") as jsonld_file:
            cycle = json.load(jsonld_file)

        with open(f"{fixtures_path}/WI5hWtsqlu/Site/2ppyB9xfAo.jsonld") as jsonld_file:
            site = json.load(jsonld_file)

        engine = Engine()

        engine.import_data(cycle, site)

        engine.data.representation_cycle['inputs']['cropResidue'] = {'value': [2.551500]}
        engine.data.representation_cycle['inputs']['cropResidue']['term'] = {'units': "kg N", 'termType': "Residues"}
        engine.data.representation_cycle['inputs']['compostAsN'] = {'term': {'@id': 'compostAsN',
                                                                             'units': 'kg N',
                                                                             'termType': 'organicFertilizer'},
                                                                    'value': [0]}

        engine.data.representation_cycle['inputs']['beefCattleSolidManureDryAsN'] = \
            {'term': {'@id': 'compostAsN',
                      'units': 'kg N',
                      'termType': 'animalProduct'
                      },
             'value': [0]
             }

        abg_node = {'term': {
                        '@id': 'aboveGroundCropResidueDryMatter',
                        'units': 'kg dry matter',
                        'termType': 'crop'},
                    'destination': {
                        '@context': "https://www-staging.hestia.earth/Term.jsonld",
                        '@id': 'leftOnField',
                        '@type': 'Term',
                        'name': 'Left on field',
                        'definition': "The product is left on the field.",
                        'termType': 'destination'
                    },
                    'properties': {
                        'nitrogenContent': {
                            'term': {
                             '@id': 'nitrogenContent',
                             '@type': 'Term'
                            },
                            'value': '0'
                        }
                    },
                    'value': [0]
                    }
        engine.data.representation_cycle['products']['aboveGroundCropResidueDryMatter'] = abg_node

        brn_node = {'term': {
                        '@id': 'belowGroundCropResidueDryMatter',
                        'units': 'kg',
                        'termType': 'crop'},
                    'destination': {
                        '@context': "https://www-staging.hestia.earth/Term.jsonld",
                        '@id': 'burning',
                        '@type': 'Term',
                        'name': 'Left on field',
                        'definition': "The product is left on the field.",
                        'termType': 'destination'
                    },
                    'properties': {
                        'nitrogenContent': {
                            'term': {
                             '@id': 'nitrogenContent',
                             '@type': 'Term'
                            },
                            'value': '0'
                        }
                    },
                    'value': [0]
                    }
        engine.data.representation_cycle['products']['belowGroundCropResidueDryMatter'] = brn_node

        self.model = NOXToAirFertilizerAndExcretaStehfestBouwman2006()
        self.representation = engine.data.representation_cycle

        self.check_expected = True
        self.noxToAirAllOrigins_expected = 0.4597373454067784
        self.noxToAirInorganicFertilizer_expected = 0.4435692463763601
        self.noxToAirOrganicFertilizer_expected = 0.0
        self.noxToAirExcreta_expected = 0.0
        self.noxToAirCropResidueDecomposition_expected = 0.0
        self.noxToAirCropResidueBurning_expected = 0.0

    def test_check_noxToAirAllOrigins(self):

        check_computed = self.model.check_noxToAirAllOrigins(self.representation)

        self.assertEqual(check_computed, self.check_expected)

    def test_check_noxToAirInorganicFertilizer(self):

        check_computed = self.model.check_noxToAirInorganicFertilizer(self.representation)

        self.assertEqual(check_computed, self.check_expected)

    def test_check_noxToAirOrganicFertilizer(self):

        check_computed = self.model.check_noxToAirOrganicFertilizer(self.representation)

        self.assertEqual(check_computed, self.check_expected)

    def test_check_noxToAirExcreta(self):

        check_computed = self.model.check_noxToAirExcreta(self.representation)

        self.assertEqual(check_computed, self.check_expected)

    def test_check_noxToAirCropResidueDecomposition(self):

        check_computed = self.model.check_noxToAirCropResidueDecomposition(self.representation)

        self.assertEqual(check_computed, self.check_expected)

    def test_check_noxToAirCropResidueBurning(self):

        check_computed = self.model.check_noxToAirCropResidueBurning(self.representation)

        self.assertEqual(check_computed, self.check_expected)

    def test_noxToAirAllOrigins(self):

        self.model.check_noxToAirAllOrigins(self.representation)
        noxToAirAllOrigins_computed = self.model.calculate_noxToAirAllOrigins()

        self.assertEqual(noxToAirAllOrigins_computed, self.noxToAirAllOrigins_expected)

    def test_noxToAirInorganicFertilizer(self):

        self.model.check_noxToAirInorganicFertilizer(self.representation)
        noxToAirInorganicFertilizer_computed = self.model.calculate_noxToAirInorganicFertilizer()

        self.assertEqual(noxToAirInorganicFertilizer_computed, self.noxToAirInorganicFertilizer_expected)

    def test_noxToAirOrganicFertilizer(self):

        self.model.check_noxToAirOrganicFertilizer(self.representation)
        noxToAirOrganicFertilizer_computed = self.model.calculate_noxToAirOrganicFertilizer()

        self.assertEqual(noxToAirOrganicFertilizer_computed, self.noxToAirOrganicFertilizer_expected)

    def test_noxToAirExcreta(self):

        self.model.check_noxToAirExcreta(self.representation)
        noxToAirExcreta_computed = self.model.calculate_noxToAirExcreta()

        self.assertEqual(noxToAirExcreta_computed, self.noxToAirExcreta_expected)

    def test_noxToAirCropResidueDecomposition(self):

        self.model.check_noxToAirCropResidueDecomposition(self.representation)
        noxToAirCropResidueDecomposition_computed = \
            self.model.calculate_noxToAirCropResidueDecomposition()

        self.assertEqual(noxToAirCropResidueDecomposition_computed, self.noxToAirCropResidueDecomposition_expected)

    def test_noxToAirCropResidueBurning(self):

        self.model.check_noxToAirCropResidueBurning(self.representation)
        noxToAirCropResidueBurning_computed = self.model.calculate_noxToAirCropResidueBurning()

        self.assertEqual(noxToAirCropResidueBurning_computed, self.noxToAirCropResidueBurning_expected)


if __name__ == '__main__':
    unittest.main()
